from torchvision import models, transforms
from torchvision.datasets import ImageFolder
from gcam import gcam
import torch
import cv2
from torch.utils.data import DataLoader
import gc
import shutil
import os
import unittest


class TestClassification(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestClassification, self).__init__(*args, **kwargs)
        self.DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.dataset = ImageFolder("data", loader=self.load_image)
        self.model = models.resnet152(pretrained=True)
        self.model.to(device=self.DEVICE)
        self.model.eval()

    def load_image(self, image_path):
        raw_image = cv2.imread(image_path)
        raw_image = cv2.resize(raw_image, (224,) * 2)
        image = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            ]
        )(raw_image[..., ::-1].copy())
        image = image.to(self.DEVICE)
        return image

    def test_gbp(self):
        model = gcam.inject(self.model, output_dir="results/resnet152/test_gbp", backend="gbp", postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=False, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        for i, batch in enumerate(data_loader):
            _ = model(batch[0])

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/resnet152"):
            shutil.rmtree("results/resnet152")

    def test_gcam(self):
        layer = 'layer4'
        model = gcam.inject(self.model, output_dir="results/resnet152/test_gcam", backend="gcam", layer=layer, postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=False, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))[0])

        for i, batch in enumerate(data_loader):
            _ = model(batch[0], label="best")

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/resnet152"):
            shutil.rmtree("results/resnet152")

    def test_ggcam(self):
        layer = 'layer4'
        model = gcam.inject(self.model, output_dir="results/resnet152/test_ggcam", backend="ggcam", layer=layer, postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=False, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))[0])

        for i, batch in enumerate(data_loader):
            _ = model(batch[0], label="best")

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/resnet152"):
            shutil.rmtree("results/resnet152")

    def test_gcampp(self):
        layer = 'layer4'
        model = gcam.inject(self.model, output_dir="results/resnet152/test_gcampp", backend="gcampp", layer=layer, postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=False, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))[0])

        for i, batch in enumerate(data_loader):
            _ = model(batch[0], label="best")

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/resnet152"):
            shutil.rmtree("results/resnet152")

if __name__ == '__main__':
    unittest.main()



import torch
from torch import nn
from gcam.backends.base import _BaseWrapper


class GuidedBackPropagation(_BaseWrapper):

    def __init__(self, model, postprocessor=None, retain_graph=False):
        """
        "Striving for Simplicity: the All Convolutional Net"
        https://arxiv.org/pdf/1412.6806.pdf
        Look at Figure 1 on page 8.
        """
        super(GuidedBackPropagation, self).__init__(model, postprocessor=postprocessor, retain_graph=retain_graph)

    def _register_hooks(self):
        """Registers the backward hooks to the layers."""
        def backward_hook(module, grad_in, grad_out):
            # Cut off negative gradients
            if isinstance(module, nn.ReLU):
                return (torch.clamp(grad_in[0], min=0.0),)

        self.remove_hook(forward=True, backward=True)
        for module in self.model.named_modules():
            self.backward_handlers.append(module[1].register_backward_hook(backward_hook))

    def forward(self, data):
        """Calls the forward() of the base."""
        self._register_hooks()
        self.data = data.requires_grad_()
        return super(GuidedBackPropagation, self).forward(self.data)

    def generate(self):
        """Generates an attention map."""
        try:
            attention_map = self.data.grad.clone()
            self.data.grad.zero_()
            B, _, *data_shape = attention_map.shape
            #attention_map = attention_map.view(B, self.channels, -1, *data_shape)
            attention_map = attention_map.view(B, 1, -1, *data_shape)
            attention_map = torch.mean(attention_map, dim=2)  # TODO: mean or sum?
            attention_map = attention_map.repeat(1, self.channels, *[1 for _ in range(self.dim)])
            attention_map = self._normalize_per_channel(attention_map)
            attention_map = attention_map.cpu().numpy()
            attention_maps = {}
            attention_maps[""] = attention_map
            return attention_maps
        except RuntimeError:
            raise RuntimeError("Number of set channels ({}) is not a multiple of the feature map channels ({})".format(self.channels, attention_map.shape[1]))



from tests.test_segmentation.unet_seg_dataset import UnetSegDataset as Dataset
from tests.test_segmentation.model.unet.unet_model import UNet
from gcam import gcam
import torch
import os
from torch.utils.data import DataLoader
import gc
import time


class Tmp():

    def __init__(self):
        self.DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.dataset = Dataset(device=self.DEVICE)
        current_path = os.path.dirname(os.path.abspath(__file__))
        CECKPOINT_PATH = os.path.join(current_path, 'model/CHECKPOINT.pth')
        self.model = UNet(n_channels=3, n_classes=1)
        self.model.load_state_dict(torch.load(CECKPOINT_PATH, map_location=self.DEVICE))
        self.model.to(device=self.DEVICE)
        self.model.eval()

    def test_gbp_hook(self):
        model = gcam.inject(self.model, output_dir="results/unet_seg/gbp", backend="gbp",
                     postprocessor="sigmoid", evaluate=False, save_scores=True, save_maps=False, save_pickle=False, metric="wioa", label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        # TODO: Memory leak finden (Oder nur beim testen?)
        start_time = time.time()
        for i, batch in enumerate(data_loader):
            #output = model(batch["img"], mask=batch["gt"])
            output = model(batch["img"])

        #model.dump()
        gc.collect()
        torch.cuda.empty_cache()
        end_time = time.time()
        print("Time elapsed: ", end_time - start_time)

        # if os.path.isdir("results"):
        #     shutil.rmtree("results")

    def test_gcam_hook(self):
        layer = 'full'
        metric = 'wioa'
        model = gcam.inject(self.model, output_dir="results/unet_seg/gcam_" + metric, backend="gcam", layer=layer, replace=False,
                                postprocessor="sigmoid", evaluate=True, save_scores=True, save_maps=False, save_pickle=False, metric=metric, label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        # TODO: Memory leak finden (Oder nur beim testen?)
        model.test_run(next(iter(data_loader))["img"])
        start_time = time.time()
        for i, batch in enumerate(data_loader):
            output = model(batch["img"], mask=batch["gt"])

        model.dump()
        end_time = time.time()
        print("Time elapsed: ", end_time - start_time)
        gc.collect()
        torch.cuda.empty_cache()

        # if os.path.isdir("results"):
        #     shutil.rmtree("results")

    def test_ggcam_hook(self):
        layer = 'full'
        metric = 'wioa'
        model = gcam.inject(self.model, output_dir="results/unet_seg/ggcam_" + metric, backend="ggcam", layer=layer,
                                postprocessor="sigmoid", evaluate=True, save_scores=True, save_maps=False, save_pickle=False, metric=metric, label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        # TODO: Memory leak finden (Oder nur beim testen?)
        model.test_run(next(iter(data_loader))["img"])
        start_time = time.time()
        for i, batch in enumerate(data_loader):
            output = model(batch["img"], mask=batch["gt"])

        model.dump()
        end_time = time.time()
        print("Time elapsed: ", end_time - start_time)
        gc.collect()
        torch.cuda.empty_cache()

        # if os.path.isdir("results"):
        #     shutil.rmtree("results")

    def test_gcampp_hook(self):
        layer = 'auto'
        metric = 'wioa'
        model = gcam.inject(self.model, output_dir="results/unet_seg/gcampp_" + metric, backend="gcampp", layer=layer,
                     postprocessor="sigmoid", evaluate=True, save_scores=True, save_maps=False, save_pickle=False, metric=metric, label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        # TODO: Memory leak finden (Oder nur beim testen?)
        model.test_run(next(iter(data_loader))["img"])
        start_time = time.time()
        for i, batch in enumerate(data_loader):
            output = model(batch["img"], mask=batch["gt"])

        model.dump()
        end_time = time.time()
        print("Time elapsed: ", end_time - start_time)
        gc.collect()
        torch.cuda.empty_cache()

        # if os.path.isdir("results"):
        #     shutil.rmtree("results")

if __name__ == '__main__':
    test = Tmp()
    #test.test_gbp_hook()
    #test.test_gcam_hook()
    #test.test_ggcam_hook()
    test.test_gcampp_hook()


from tests.test_segmentation.unet_seg_dataset import UnetSegDataset as Dataset
from tests.test_segmentation.model.unet.unet_model import UNet
from gcam import gcam
import torch
import os
import unittest
from torch.utils.data import DataLoader
import gc
import shutil


class TestSegmentation(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestSegmentation, self).__init__(*args, **kwargs)
        self.DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.dataset = Dataset(device=self.DEVICE)
        current_path = os.path.dirname(os.path.abspath(__file__))
        CECKPOINT_PATH = os.path.join(current_path, 'model/CHECKPOINT.pth')
        self.model = UNet(n_channels=3, n_classes=1)
        self.model.load_state_dict(torch.load(CECKPOINT_PATH, map_location=self.DEVICE))
        self.model.to(device=self.DEVICE)
        self.model.eval()

    def test_gbp(self):
        model = gcam.inject(self.model, output_dir="results/unet_seg/gbp", backend="gbp",
                     postprocessor="sigmoid", evaluate=True, save_scores=True, save_maps=False, save_pickle=False, metric="wioa", label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)

        for i, batch in enumerate(data_loader):
            _ = model(batch["img"], mask=batch["gt"])

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/unet_seg"):
            shutil.rmtree("results/unet_seg")

    def test_gcam(self):
        layer = 'full'
        metric = 'wioa'
        model = gcam.inject(self.model, output_dir="results/unet_seg/gcam_" + metric, backend="gcam", layer=layer,
                                postprocessor="sigmoid", evaluate=True, save_scores=True, save_maps=False, save_pickle=False, metric=metric, label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))["img"])

        for i, batch in enumerate(data_loader):
            _ = model(batch["img"], mask=batch["gt"])

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/unet_seg"):
            shutil.rmtree("results/unet_seg")

    def test_ggcam(self):
        layer = 'full'
        metric = 'wioa'
        model = gcam.inject(self.model, output_dir="results/unet_seg/ggcam_" + metric, backend="ggcam", layer=layer,
                                postprocessor="sigmoid", evaluate=True, save_scores=True, save_maps=False, save_pickle=False, metric=metric, label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))["img"])

        for i, batch in enumerate(data_loader):
            _ = model(batch["img"], mask=batch["gt"])

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/unet_seg"):
            shutil.rmtree("results/unet_seg")

    def test_gcampp(self):
        layer = 'full'
        metric = 'wioa'
        model = gcam.inject(self.model, output_dir="results/unet_seg/gcampp_" + metric, backend="gcampp", layer=layer,
                     postprocessor="sigmoid", evaluate=True, save_scores=True, save_maps=False, save_pickle=False, metric=metric, label=lambda x: 0.5 < x, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))["img"])

        for i, batch in enumerate(data_loader):
            _ = model(batch["img"], mask=batch["gt"])

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results/unet_seg"):
            shutil.rmtree("results/unet_seg")


if __name__ == '__main__':
    unittest.main()



from torchvision import models, transforms
from torchvision.datasets import ImageFolder
from gcam import gcam
import torch
import cv2
from torch.utils.data import DataLoader
import gc
import shutil
import os


class Tmp():

    def __init__(self):
        self.DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.dataset = ImageFolder("data", loader=self.load_image)
        self.model = models.resnet152(pretrained=True)
        self.model.to(device=self.DEVICE)
        self.model.eval()

    def load_image(self, image_path):
        raw_image = cv2.imread(image_path)
        raw_image = cv2.resize(raw_image, (224,) * 2)
        image = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            ]
        )(raw_image[..., ::-1].copy())
        image = image.to(self.DEVICE)
        return image

    def test_gbp(self):
        model = gcam.inject(self.model, output_dir="results/resnet152/test_gbp", backend="gbp", postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=True, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        for i, batch in enumerate(data_loader):
            _ = model(batch[0])

        model.dump()

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results"):
            shutil.rmtree("results")

    def test_gcam(self):
        layer = 'layer4'
        model = gcam.inject(self.model, output_dir="results/resnet152/test_gcam", backend="gcam", layer=layer, postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=True, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))[0])

        for i, batch in enumerate(data_loader):
            _ = model(batch[0], label="best")

        model.dump()

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results"):
            shutil.rmtree("results")

    def test_ggcam(self):
        layer = 'layer4'
        model = gcam.inject(self.model, output_dir="results/resnet152/test_ggcam", backend="ggcam", layer=layer, postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=True, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))[0])

        for i, batch in enumerate(data_loader):
            _ = model(batch[0], label="best")

        model.dump()

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results"):
            shutil.rmtree("results")

    def test_gcampp(self):
        layer = 'layer4'
        model = gcam.inject(self.model, output_dir="results/resnet152/test_gcampp", backend="gcampp", layer=layer, postprocessor="softmax",
                    evaluate=False, save_scores=False, save_maps=True, save_pickle=False, channels=1)
        model.eval()
        data_loader = DataLoader(self.dataset, batch_size=1, shuffle=False)
        model.test_run(next(iter(data_loader))[0])

        for i, batch in enumerate(data_loader):
            _ = model(batch[0], label="best")

        model.dump()

        del model
        gc.collect()
        torch.cuda.empty_cache()

        if os.path.isdir("results"):
            shutil.rmtree("results")

if __name__ == '__main__':
    test = Tmp()
    #test.test_gbp()
    #test.test_gcam()
    #test.test_ggcam()
    test.test_gcampp()
